/*
 * Written by Manuel Mausz <manuel (at) mausz (dot) at>
 * Some code was borrowed from pam_userdb written by Cristian Gafton <gafton@redhat.com>
 */

#include "config.h"
#include "pam_useralias.h"

#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <fcntl.h>
#include <ctype.h>

#ifdef HAVE_NDBM_H
# include <ndbm.h>
#else
# ifdef HAVE_DB_H
#  define DB_DBM_HSEARCH 1  /* use the dbm interface */
#  define HAVE_DBM          /* for BerkDB 5.0 and later */
#  include <db.h>
# else
#  error "failed to find a libdb or equivalent"
# endif
#endif

#include <security/pam_modules.h>
#include <security/pam_appl.h>
#include <security/pam_ext.h>

static int
_pam_parse(pam_handle_t *pamh, int argc, const char **argv,
    const char **database)
{
  int ctrl;

  *database = NULL;

  /* step through arguments */
  for (ctrl = 0; argc-- > 0; ++argv)
  {
    /* generic options */
    if (!strcmp(*argv, "debug"))
      ctrl |= PAM_DEBUG_ARG;
    else if (!strcasecmp(*argv, "dump"))
      ctrl |= PAM_DUMP_ARG;
    else if (!strcasecmp(*argv, "tolower"))
      ctrl |= PAM_TOLOWER_ARG;
    else if (!strncasecmp(*argv, "db=", 3))
    {
      *database = (*argv) + 3;
      if (**database == '\0')
      {
        *database = NULL;
        pam_syslog(pamh, LOG_ERR, "db= specification missing argument - ignored");
      }
    }
    else
      pam_syslog(pamh, LOG_ERR, "unknown option: %s", *argv);
  }

  return ctrl;
}

/*
 * looks up an username (alias) in a database (as key)
 * and returns the realuser (value)
 *
 * return values:
 *    1  = user not found
 *    0  = user found
 *   -1  = system error
 */
static int
user_lookup(pam_handle_t *pamh, const char *database, const char *username,
    const char **realuser, int ctrl)
{
  DBM *dbm;
  datum key, data;
  register unsigned int i;

  *realuser = NULL;

  if (ctrl & PAM_DEBUG_ARG)
    pam_syslog(pamh, LOG_INFO, "searching for alias '%s' in database", username);

  /* open the DB file. */
  dbm = dbm_open(database, O_RDONLY, 0644);
  if (dbm == NULL)
  {
    pam_syslog(pamh, LOG_ERR, "user_lookup: could not open database '%s': %m", database);
    return -1;
  }

  /* dump out the database contents for debugging */
  if (ctrl & PAM_DUMP_ARG)
  {
    pam_syslog(pamh, LOG_INFO, "Database dump:");
    for (key = dbm_firstkey(dbm); key.dptr != NULL; key = dbm_nextkey(dbm))
    {
      data = dbm_fetch(dbm, key);
      pam_syslog(pamh, LOG_INFO, "key[len=%d] = '%.*s', data[len=%d] = '%.*s'",
          key.dsize, key.dsize, key.dptr, data.dsize, data.dsize, data.dptr);
    }
  }

  /* do some more init work */
  memset(&key, 0, sizeof(key));
  memset(&data, 0, sizeof(data));
  key.dptr = strdup(username);
  key.dsize = strlen(username);

  if (ctrl & PAM_TOLOWER_ARG)
  {
    for (i = 0; i < key.dsize; ++i)
      key.dptr[i] = tolower(key.dptr[i]);
    if (ctrl & PAM_DEBUG_ARG)
      pam_syslog(pamh, LOG_INFO, "tolower-param is enabled. lowered alias is: '%s'", key.dptr);
  }

  data = dbm_fetch(dbm, key);
  free(key.dptr);

  if (data.dptr != NULL)
    *realuser = strndup(data.dptr, data.dsize);

  dbm_close(dbm);

  return (*realuser == NULL) ? 1 : 0;
}

PAM_EXTERN int
pam_sm_authenticate(pam_handle_t *pamh, int flags UNUSED,
    int argc, const char **argv)
{
  const char *username;
  const char *database = NULL;
  const char *realuser = NULL;
  int retval = PAM_AUTH_ERR, ctrl;

  /* parse arguments */
  ctrl = _pam_parse(pamh, argc, argv, &database);
  if (database == NULL)
  {
     pam_syslog(pamh, LOG_ERR, "unable to get the database name");
     return PAM_SERVICE_ERR;
  }

  /* get the username (aka alias) */
  retval = pam_get_user(pamh, &username, NULL);
  if ((retval != PAM_SUCCESS) || !username)
  {
     pam_syslog(pamh, LOG_WARNING, "unable to get the username");
     return PAM_USER_UNKNOWN;
  }

  /* now use the username to look up the real systemuser in the database file */
  retval = user_lookup(pamh, database, username, &realuser, ctrl);
  switch (retval)
  {
    case -1:
       /* some sort of system error. the log was already printed */
      return PAM_SERVICE_ERR;
    case 1:
      /* the user does not exist in the database */
      if (ctrl & PAM_DEBUG_ARG)
        pam_syslog(pamh, LOG_NOTICE, "alias '%s' not found in the database", username);
      return PAM_USER_UNKNOWN;
    case 0:
      /* the user does exist in the database */
      pam_syslog(pamh, LOG_NOTICE, "alias '%s' found for system user '%s'", username, realuser);
      if (pam_set_item(pamh, PAM_USER, realuser) != PAM_SUCCESS)
        pam_syslog(pamh, LOG_WARNING, "unable to change username from '%s' to '%s'", username, realuser);
      return PAM_SUCCESS;
    default:
      /* we don't know anything about this return value */
      pam_syslog(pamh, LOG_ERR, "internal module error (retval=%d, alias='%s'", retval, username);
      return PAM_SERVICE_ERR;
  }

  /* should not be reached */
  return PAM_IGNORE;
}

PAM_EXTERN int
pam_sm_setcred(pam_handle_t *pamh UNUSED, int flags UNUSED,
    int argc UNUSED, const char **argv UNUSED)
{
  return PAM_SUCCESS;
}

#ifdef PAM_STATIC

/* static module data */
struct pam_module _pam_useralias_modstruct = {
  MODULE_NAME,
  pam_sm_authenticate,
  pam_sm_setcred,
  NULL,
  NULL,
  NULL,
  NULL,
};

#endif
